/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/posix/userdata.c: User data access routines for
 * POSIX-compatible systems.
 */

/*
 * This file implements the common portion of user data access for systems
 * which use the standard POSIX filesystem interface for storing user data.
 * The system-specific code need only implement sys_userdata_get_data_path();
 * the routines in this file will call that function to retrieve the path
 * for accessing files (which may change over the life of the program,
 * though the value at sys_userdata_perform() time for a particular
 * operation will be used for that operation regardless of any concurrent
 * changes).
 *
 * The default file pathnames for this implementation are as follows, where
 * <path> is the path returned by sys_userdata_get_data_path():
 *
 *    - Save files: <path>save/save-NNNN.{bin,png}
 *         (NNNN is the save number, zero-padded to 4 digits)
 *    - Settings file: <path>settings.bin
 *    - Per-user statistics file: <path>stats.bin
 *    - Arbitrary data files: <path><datafile-path>
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/posix/fileutil.h"
#include "src/sysdep/posix/path_max.h"
#include "src/sysdep/posix/userdata.h"
#include "src/userdata.h"
#include "src/utility/png.h"

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>

/*************************************************************************/
/****************** Global data (only used for testing) ******************/
/*************************************************************************/

const char *
#ifndef INCLUDE_TESTS
    const
#endif
    TEST_posix_userdata_path = NULL;

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Table of registered per-user statistics and current values. */
static struct {
    UserStatInfo info;
    double value;
} *stats;
static int num_stats;

/* Flag indicating whether we have unsaved changes to statistics data
 * (i.e., if we updated statistics but failed to save the updates). */
static uint8_t stats_update_needed;

/* Mutex for accessing the stats array. */
static SysMutexID stats_mutex;

/*-----------------------------------------------------------------------*/

/* Local routine declarations. */

/*-------- Helper functions --------*/

/**
 * generate_path:  Generate the pathname to use for the given operation
 * (assuming no override_path has been specified).
 *
 * [Parameters]
 *     params: Parameter block for operation.
 *     buffer: Buffer into which to place generated pathname.
 *     bufsize: Size of buffer, in bytes.
 * [Return value]
 *     True if the pathname was successfully generated, false on buffer
 *     overflow or other error.
 */
static int generate_path(const SysUserDataParams *params,
                         char *buffer, int bufsize);

/**
 * generate_save_screenshot_path:  Generate the pathname to use for the
 * screenshot associated with the given save file pathname.
 *
 * [Parameters]
 *     path: Save file pathname.
 *     buffer: Buffer into which to place generated pathname.
 *     bufsize: Size of buffer, in bytes.
 * [Return value]
 *     True if the pathname was successfully generated, false on buffer
 *     overflow or other error.
 */
static int generate_save_screenshot_path(const char *path,
                                         char *buffer, int bufsize);

/**
 * save_stats_file:  Save the current statistics data in the local
 * statistics file.  Helper function for do_stats_update().
 *
 * [Parameters]
 *     path: Pathname of file to write.
 * [Return value]
 *     True on success, false on failure.
 */
static int save_stats_file(const char *path);

/*-------- Operation-specific handling --------*/

/**
 * do_save:  Perform a generic save operation.
 *
 * [Parameters]
 *     params: Parameter block for operation.
 *     path: Pathname of file to operate on.
 */
static int do_save(SysUserDataParams *params, const char *path);

/**
 * do_load:  Perform a generic load operation.
 *
 * [Parameters]
 *     params: Parameter block for operation.
 *     path: Pathname of file to operate on.
 */
static int do_load(SysUserDataParams *params, const char *path);

/**
 * do_delete:  Perform a generic delete operation.
 *
 * [Parameters]
 *     params: Parameter block for operation.
 *     path: Pathname of file to operate on.
 */
static int do_delete(SysUserDataParams *params, const char *path);

/**
 * do_scan_savefiles:  Perform a SCAN_SAVEFILES operation.
 *
 * [Parameters]
 *     params: Parameter block for operation.
 *     path: Pathname of directory to scan.
 */
static int do_scan_savefiles(SysUserDataParams *params, const char *path);

/**
 * do_save_image:  Perform a save operation for an image file.
 *
 * [Parameters]
 *     params: Parameter block for operation.
 *     path: Pathname of file to operate on.
 */
static int do_save_image(SysUserDataParams *params, const char *path);

/**
 * do_load_image:  Perform a load operation for an image file.
 *
 * [Parameters]
 *     params: Parameter block for operation.
 *     path: Pathname of file to operate on.
 */
static int do_load_image(SysUserDataParams *params, const char *path);

/**
 * do_stats_register:  Perform a REGISTER_STATS operation.  On entry, the
 * stats array has been allocated to hold num_stats entries, and the mutex
 * is already locked.
 *
 * [Parameters]
 *     params: Parameter block for operation.
 *     path: Pathname of file to operate on.
 */
static int do_stats_register(SysUserDataParams *params, const char *path);

/**
 * do_stats_update:  Perform an UPDATE_STATS operation.  On entry, the
 * stats array has been initialized with num_stats entries.
 *
 * [Parameters]
 *     params: Parameter block for operation.
 *     path: Pathname of file to operate on.
 */
static int do_stats_update(SysUserDataParams *params, const char *path);

/**
 * do_stats_retrieve:  Perform a RETRIEVE_STATS operation.  On entry, the
 * stats array has been initialized with num_stats entries.
 *
 * [Parameters]
 *     params: Parameter block for operation.
 *     path: Pathname of file to operate on.
 */
static int do_stats_retrieve(SysUserDataParams *params, const char *path);

/**
 * do_stats_clear:  Perform a CLEAR_STATS operation.  On entry, the stats
 * array has been initialized with num_stats entries.
 *
 * [Parameters]
 *     params: Parameter block for operation.
 *     path: Pathname of file to operate on.
 */
static int do_stats_clear(SysUserDataParams *params, const char *path);

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

int sys_userdata_perform(SysUserDataParams *params)
{
    PRECOND(params != NULL, return 0);

    /* Generate the pathname for this operation (unless an override path
     * was specified). */

    char pathbuf[PATH_MAX];

    if (params->override_path) {
        if (TEST_posix_userdata_path) {
            if (!strformat_check(pathbuf, sizeof(pathbuf), "%s%s",
                                 TEST_posix_userdata_path,
                                 params->override_path)) {
                DLOG("Buffer overflow on path: %s%s",
                     TEST_posix_userdata_path, params->override_path);
                return 0;
            }
        } else if (!strformat_check(pathbuf, sizeof(pathbuf), "%s",
                                    params->override_path)) {
            DLOG("Buffer overflow on path: %s", params->override_path);
            return 0;
        }
    } else {
        if (!generate_path(params, pathbuf, sizeof(pathbuf))) {
            return 0;
        }
    }

    /* For REGISTER_STATS, (re)allocate the statistics array first. */

    if (params->operation == SYS_USERDATA_REGISTER_STATS) {
        sys_mutex_lock(stats_mutex);
        if (stats) {
            mem_free(stats);
            stats = NULL;
            num_stats = 0;
        }
        stats = mem_alloc(sizeof(*stats) * params->stat_info_len, 0, 0);
        if (UNLIKELY(!stats)) {
            DLOG("No memory for %d statistic entries", params->stat_info_len);
            sys_mutex_unlock(stats_mutex);
            return 0;
        }
        num_stats = params->stat_info_len;
        /* Leave the mutex locked for do_stats_register(). */
    }

    /* Perform the actual operation and return. */

    int result = -1;
    switch (params->operation) {
      case SYS_USERDATA_SAVE_SAVEFILE:
      case SYS_USERDATA_SAVE_SETTINGS:
      case SYS_USERDATA_SAVE_DATA:
        result = do_save(params, pathbuf);
        break;

      case SYS_USERDATA_LOAD_SAVEFILE:
      case SYS_USERDATA_LOAD_SETTINGS:
      case SYS_USERDATA_LOAD_DATA:
        result = do_load(params, pathbuf);
        break;

      case SYS_USERDATA_DELETE_SAVEFILE:
      case SYS_USERDATA_DELETE_DATA:
        result = do_delete(params, pathbuf);
        break;

      case SYS_USERDATA_SCAN_SAVEFILES:
        result = do_scan_savefiles(params, pathbuf);
        break;

      case SYS_USERDATA_SAVE_SCREENSHOT:
        result = do_save_image(params, pathbuf);
        break;

      case SYS_USERDATA_REGISTER_STATS:
        /* Mutex is already locked. */
        result = do_stats_register(params, pathbuf);
        if (!result) {
            mem_free(stats);
            stats = NULL;
            num_stats = 0;
        }
        sys_mutex_unlock(stats_mutex);
        break;

      case SYS_USERDATA_UPDATE_STATS:
        sys_mutex_lock(stats_mutex);
        result = do_stats_update(params, pathbuf);
        sys_mutex_unlock(stats_mutex);
        break;

      case SYS_USERDATA_RETRIEVE_STATS:
        sys_mutex_lock(stats_mutex);
        result = do_stats_retrieve(params, pathbuf);
        sys_mutex_unlock(stats_mutex);
        break;

      case SYS_USERDATA_CLEAR_STATS:
        sys_mutex_lock(stats_mutex);
        result = do_stats_clear(params, pathbuf);
        sys_mutex_unlock(stats_mutex);
        break;
    }

    ASSERT(result >= 0 || !"Operation code was invalid.", result = 0);

    return result;
}

/*************************************************************************/
/********************* Exported test helper routines *********************/
/*************************************************************************/

#ifdef INCLUDE_TESTS

void TEST_posix_userdata_unregister_stats(void)
{
    mem_free(stats);
    stats = NULL;
    num_stats = 0;
}

#endif

/*************************************************************************/
/*********************** Library-internal routines ***********************/
/*************************************************************************/

int posix_userdata_init(void)
{
    stats_mutex = sys_mutex_create(0);
    if (UNLIKELY(!stats_mutex)) {
        DLOG("Failed to create mutex for stats array");
        return 0;
    }
    return 1;
}

/*************************************************************************/
/******************* Local routines: Helper functions ********************/
/*************************************************************************/

static int generate_path(const SysUserDataParams *params,
                         char *buffer, int bufsize)
{
    PRECOND(params != NULL, return 0);
    PRECOND(buffer != NULL, return 0);
    PRECOND(bufsize > 0, return 0);

    int path_len;

    /* Retrieve the base user directory path from system-specific code. */

    char *sys_path = TEST_posix_userdata_path
        ? mem_strdup(TEST_posix_userdata_path, MEM_ALLOC_TEMP)
        : sys_userdata_get_data_path(params->program_name);
    if (!sys_path) {
        DLOG("Failed to get base path");
        return 0;
    }
    path_len = strformat(buffer, bufsize, "%s", sys_path);
    mem_free(sys_path);
    if (path_len >= bufsize) {
        DLOG("Path buffer overflow on user data directory");
        return 0;
    }

    /* Append the operation-specific directory and filename. */

    /* Record whether the operation code is valid.  (We use this variable
     * instead of a "default" case in the switch() so the compiler can
     * detect any missing cases.) */
    int operation_valid = 0;

    switch (params->operation) {

      case SYS_USERDATA_SAVE_SAVEFILE:
      case SYS_USERDATA_LOAD_SAVEFILE:
      case SYS_USERDATA_DELETE_SAVEFILE:
        path_len += strformat(buffer + path_len, bufsize - path_len,
                              "save/save-%04d.bin", params->savefile_num);
        operation_valid = 1;
        break;

      case SYS_USERDATA_SCAN_SAVEFILES:
        /* Pass back the path of the directory to read. */
        path_len += strformat(buffer + path_len, bufsize - path_len, "save");
        operation_valid = 1;
        break;

      case SYS_USERDATA_SAVE_SETTINGS:
      case SYS_USERDATA_LOAD_SETTINGS:
        path_len += strformat(buffer + path_len, bufsize - path_len,
                              "settings.bin");
        operation_valid = 1;
        break;

      case SYS_USERDATA_SAVE_SCREENSHOT:
      {
        const int base_path_len =
            path_len + strformat(buffer + path_len, bufsize - path_len,
                                 "screenshots/screen");
        int filenum = 0;
        do {
            path_len = base_path_len + strformat(buffer + base_path_len,
                                                 bufsize - base_path_len,
                                                 "%d.png", filenum);
            filenum++;
        } while (access(buffer, F_OK) == 0);
        operation_valid = 1;
        break;
      }

      case SYS_USERDATA_SAVE_DATA:
      case SYS_USERDATA_LOAD_DATA:
      case SYS_USERDATA_DELETE_DATA:
        ASSERT(params->datafile_path != NULL, return 0);
        path_len += strformat(buffer + path_len, bufsize - path_len,
                              "%s", params->datafile_path);
        operation_valid = 1;
        break;

      case SYS_USERDATA_REGISTER_STATS:
      case SYS_USERDATA_UPDATE_STATS:
      case SYS_USERDATA_RETRIEVE_STATS:
      case SYS_USERDATA_CLEAR_STATS:
        path_len += strformat(buffer + path_len, bufsize - path_len,
                              "stats.bin");
        operation_valid = 1;
        break;

    }  // switch (params->operation)

    ASSERT(operation_valid, return 0);

    if (path_len >= bufsize) {
        DLOG("Path buffer overflow on user data file");
        return 0;
    }

    /* Pathname successfully generated. */

    return 1;
}

/*-----------------------------------------------------------------------*/

static int generate_save_screenshot_path(const char *path,
                                         char *buffer, int bufsize)
{
    PRECOND(path != NULL, return 0);
    PRECOND(buffer != NULL, return 0);
    PRECOND(bufsize > 0, return 0);

    const char *slash = strrchr(path, '/');
    const char *dot = strrchr(path, '.');
    if (!dot || (slash && dot < slash)) {
        dot = path + strlen(path);
    }
    if (!strformat_check(buffer, bufsize, "%.*s.png", (int)(dot-path), path)) {
        DLOG("Buffer overflow on screenshot pathname");
        return 0;
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

static int save_stats_file(const char *path)
{
    /* Figure out how much buffer space we need. */
    int save_size = 0;
    for (int i = 0; i < num_stats; i++) {
        switch (stats[i].info.type) {
          case USERDATA_STAT_FLAG:
            save_size += 1;
            break;
          case USERDATA_STAT_UINT32:
          case USERDATA_STAT_UINT32_MAX:
            save_size += 4;
            break;
          case USERDATA_STAT_DOUBLE:
          case USERDATA_STAT_DOUBLE_MAX:
            save_size += 8;
            break;
        }
    }

    /* Create the file data in a memory buffer. */
    char *save_buffer = mem_alloc(save_size, 0, MEM_ALLOC_TEMP);
    if (!save_buffer) {
        DLOG("Out of memory for statistics data (%d bytes)", save_size);
        return 0;
    }
    char *data = save_buffer;
    for (int i = 0; i < num_stats; i++) {
        switch (stats[i].info.type) {
          case USERDATA_STAT_FLAG:
            *data++ = (stats[i].value != 0) ? 1 : 0;
            break;
          case USERDATA_STAT_UINT32:
          case USERDATA_STAT_UINT32_MAX:
            ASSERT(stats[i].value >= 0, stats[i].value = 0);
            ASSERT(stats[i].value <= (double)0xFFFFFFFFU,
                   stats[i].value = (double)0xFFFFFFFFU);
            {
                uint32_t value = (uint32_t)stats[i].value;
                data[0] = value>>24 & 0xFF;
                data[1] = value>>16 & 0xFF;
                data[2] = value>> 8 & 0xFF;
                data[3] = value>> 0 & 0xFF;
            }
            data += 4;
            break;
          case USERDATA_STAT_DOUBLE:
          case USERDATA_STAT_DOUBLE_MAX:
            {
                union {uint64_t i; double d;} u;
                u.d = stats[i].value;
                data[0] = u.i>>56 & 0xFF;
                data[1] = u.i>>48 & 0xFF;
                data[2] = u.i>>40 & 0xFF;
                data[3] = u.i>>32 & 0xFF;
                data[4] = u.i>>24 & 0xFF;
                data[5] = u.i>>16 & 0xFF;
                data[6] = u.i>> 8 & 0xFF;
                data[7] = u.i>> 0 & 0xFF;
            }
            data += 8;
            break;
        }  // switch (stats[i].info.type)
    }  // for (int i = 0; i < num_stats; i++)

    /* Write out the data. */
    const int result = posix_write_file(path, save_buffer, save_size);
    mem_free(save_buffer);
    if (!result) {
        DLOG("Failed to save statistics file");
        return 0;
    }

    return 1;
}

/*************************************************************************/
/************** Local routines: Operation-specific handling **************/
/*************************************************************************/

static int do_save(SysUserDataParams *params, const char *path)
{
    PRECOND(params != NULL, return 0);
    PRECOND(path != NULL, return 0);

    const char *data = params->save_data;
    uint32_t size = params->save_size;

    /* Write the file to disk. */

    int ok = posix_write_file(path, data, size);
    if (!ok) {
        goto fail;
    }

    /* If this is a save file operation, save the screenshot separately,
     * or remove any existing screenshot if none was given with this
     * operation.  Note that we don't treat failure here as a failure of
     * the whole operation, since the save itself has already succeeded.
     * (Normally this will never fail, since posix_write_file() requires
     * write access to the directory, thus ensuring that a second write or
     * a remove() call will succeed.  It could potentially fail if the
     * filesystem becomes full while writing the screenshot.) */

    if (params->operation == SYS_USERDATA_SAVE_SAVEFILE) {
        char png_path[PATH_MAX];
        if (generate_save_screenshot_path(path, png_path, sizeof(png_path))) {
            if (params->save_image) {
                do_save_image(params, png_path);
            } else {
                if (remove(png_path) != 0 && errno != ENOENT) {
                    DLOG("Warning: remove(%s) failed: %s", png_path,
                         strerror(errno));
                }
            }
        }
    }

    /* Success! */

    return 1;

    /* Error cases jump down here. */

  fail:
    return 0;
}

/*-----------------------------------------------------------------------*/

static int do_load(SysUserDataParams *params, const char *path)
{
    PRECOND(params != NULL, return 0);
    PRECOND(path != NULL, return 0);

    char *buffer = NULL;
    uint32_t size = 0;

    /* Load the file into memory. */

    ssize_t size_temp;
    buffer = posix_read_file(path, &size_temp, 0);
    if (!buffer) {
        goto fail;
    }
#if SIZE_MAX > 0xFFFFFFFFUL
    if ((size_t)size_temp > 0xFFFFFFFFUL) {
        /* The user has probably given up waiting for the load long ago,
         * but let's avoid overflow anyway. */
        DLOG("%s: File too large", path);
        mem_free(buffer);
        goto fail;
    }
#endif
    size = (uint32_t)size_temp;

    /* If this is a save file operation, try to load the screenshot as well. */

    if (params->operation == SYS_USERDATA_LOAD_SAVEFILE) {
        char png_path[PATH_MAX];
        if (generate_save_screenshot_path(path, png_path, sizeof(png_path))) {
            const int result = do_load_image(params, png_path);
            if (!result) {
                params->load_image = NULL;
                params->load_image_width = 0;
                params->load_image_height = 0;
            }
        }
    }

    /* Success! */

    params->load_data = buffer;
    params->load_size = size;
    return 1;

    /* Error cases jump down here. */

  fail:
    mem_free(buffer);
    return 0;
}

/*-----------------------------------------------------------------------*/

static int do_delete(SysUserDataParams *params, const char *path)
{
    PRECOND(params != NULL, return 0);
    PRECOND(path != NULL, return 0);

    if (remove(path) != 0 && errno != ENOENT) {
        DLOG("remove(%s) failed: %s", path, strerror(errno));
        return 0;
    }

    if (params->operation == SYS_USERDATA_DELETE_SAVEFILE) {
        char png_path[PATH_MAX];
        if (generate_save_screenshot_path(path, png_path, sizeof(png_path))) {
            if (remove(png_path) != 0 && errno != ENOENT) {
                /* As with do_save(), we don't treat this as a failure. */
                DLOG("Warning: remove(%s) failed: %s", png_path,
                     strerror(errno));
            }
        }
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

static int do_scan_savefiles(SysUserDataParams *params, const char *path)
{
    PRECOND(params != NULL, return 0);
    PRECOND(params->scan_buffer != NULL, return 0);
    PRECOND(params->scan_count > 0, return 0);
    PRECOND(path != NULL, return 0);

    mem_clear(params->scan_buffer, params->scan_count);

    SysDir *dir = sys_dir_open(path);
    if (!dir) {
        /* If the directory is known not to exist, we can safely return
         * success with an array of zeroes.  Otherwise, return failure. */
        return (sys_last_error() == SYSERR_FILE_NOT_FOUND);
    }

    const char *filename;
    while ((filename = sys_dir_read(dir)) != NULL) {
        if (strncmp(filename, "save-", 5) != 0) {
            continue;
        }
        const int i = 5 + strspn(filename+5, "0123456789");
        if (strcmp(filename+i, ".bin") != 0) {
            continue;
        }
        const int num = atoi(filename+5);
        if (num >= params->savefile_num) {
            const int offset = num - params->savefile_num;
            if (offset < params->scan_count) {
                params->scan_buffer[offset] = 1;
            }
        }
    }

    sys_dir_close(dir);
    return 1;
}

/*-----------------------------------------------------------------------*/

static int do_save_image(SysUserDataParams *params, const char *path)
{
    PRECOND(params != NULL, return 0);
    PRECOND(path != NULL, return 0);

    uint32_t png_size;
    void *png_data = png_create(params->save_image,
                                params->save_image_width,
                                params->save_image_height,
                                0, -1, 0, MEM_ALLOC_TEMP, &png_size);
    if (!png_data) {
        DLOG("Failed to generate PNG file for screenshot");
        return 0;
    }
    const int result = posix_write_file(path, png_data, png_size);
    mem_free(png_data);
    return result;
}

/*-----------------------------------------------------------------------*/

static int do_load_image(SysUserDataParams *params, const char *path)
{
    PRECOND(params != NULL, return 0);
    PRECOND(path != NULL, return 0);

    ssize_t png_size;
    void *png_data = posix_read_file(path, &png_size, MEM_ALLOC_TEMP);
    if (!png_data) {
        goto fail;
    }
#if SIZE_MAX > 0xFFFFFFFFUL
    if ((size_t)png_size > 0xFFFFFFFFU) {
        DLOG("%s: File too large", path);
        mem_free(png_data);
        goto fail;
    }
#endif

    params->load_image = png_parse(png_data, (uint32_t)png_size, 0,
                                   &params->load_image_width,
                                   &params->load_image_height);
    mem_free(png_data);
    if (!params->load_image) {
        DLOG("Failed to parse PNG file %s", path);
        goto fail;
    }

    return 1;

  fail:
    return 0;
}

/*-----------------------------------------------------------------------*/

static int do_stats_register(SysUserDataParams *params, const char *path)
{
    PRECOND(params != NULL, return 0);
    PRECOND(path != NULL, return 0);
    PRECOND(stats != NULL, return 0);

    /* Copy the caller's statistic list into our array, ensuring that
     * there are no duplicate IDs or invalid types. */

    int num_registered = 0;
    int failed = 0;
    for (int i = 0; i < params->stat_info_len; i++) {

        int j;
        for (j = 0; j < num_registered; j++) {
            if (stats[j].info.id == params->stat_info[i].id) {
                break;
            }
        }
        if (j < num_registered) {
            DLOG("Stats entry %d: duplicate ID %u", i, stats[j].info.id);
            failed = 1;
            continue;
        }

        /* Let the compiler help us make sure we don't miss any types. */
        int type_ok = 0;
        switch (params->stat_info[i].type) {
          case USERDATA_STAT_FLAG:
          case USERDATA_STAT_UINT32:
          case USERDATA_STAT_DOUBLE:
          case USERDATA_STAT_UINT32_MAX:
          case USERDATA_STAT_DOUBLE_MAX:
            type_ok = 1;
            break;
        }
        if (!type_ok) {
            DLOG("Stats entry %d (ID %u): invalid type %d",
                 i, params->stat_info[i].id, params->stat_info[i].type);
            failed = 1;
            continue;
        }

        ASSERT(num_registered <= num_stats, failed = 1; break);
        stats[num_registered].info = params->stat_info[i];
        stats[num_registered].value = 0;
        num_registered++;

    }  // for (int i = 0; i < params->stat_info_len; i++)

    if (failed) {
        return 0;
    }

    num_stats = num_registered;  // These should be the same, but just in case.

    /* Load this user's statistics. */

    if (access(path, F_OK) != 0) {
        /* If the file doesn't exist, don't treat that as an error --
         * just leave the default values in place. */
        return 1;
    }

    ssize_t size;
    uint8_t * const data = posix_read_file(path, &size, MEM_ALLOC_TEMP);
    if (!data) {
        DLOG("Failed to load statistics file");
        return 0;
    }

    ssize_t pos = 0;
    for (int i = 0; i < num_stats; i++) {
        switch (stats[i].info.type) {

          case USERDATA_STAT_FLAG:
            if (pos+1 > size) {
                DLOG("Missing data in statistics file");
                goto exit;
            }
            if (data[pos] != 0 && data[pos] != 1) {
                DLOG("Invalid data in statistics file (ID %u)",
                     stats[i].info.id);
            } else {
                stats[i].value = data[pos++];
            }
            break;

          case USERDATA_STAT_UINT32:
          case USERDATA_STAT_UINT32_MAX:
            if (pos+4 > size) {
                DLOG("Missing data in statistics file");
                goto exit;
            }
            stats[i].value = data[pos+0]<<24
                           | data[pos+1]<<16
                           | data[pos+2]<< 8
                           | data[pos+3]<< 0;
            pos += 4;
            break;

          case USERDATA_STAT_DOUBLE:
          case USERDATA_STAT_DOUBLE_MAX:
            if (pos+8 > size) {
                DLOG("Missing data in statistics file");
                goto exit;
            }
            {
                union {uint64_t i; double d;} u;
                u.i = (uint64_t)data[pos+0]<<56
                    | (uint64_t)data[pos+1]<<48
                    | (uint64_t)data[pos+2]<<40
                    | (uint64_t)data[pos+3]<<32
                    | (uint64_t)data[pos+4]<<24
                    | (uint64_t)data[pos+5]<<16
                    | (uint64_t)data[pos+6]<< 8
                    | (uint64_t)data[pos+7]<< 0;
                stats[i].value = u.d;
            }
            pos += 8;
            break;

        }  // switch (stats[i].info.type)
    }  // for (int i = 0; i < num_stats; i++)

  exit:
    mem_free(data);
    params->load_data = NULL;
    params->load_size = 0;
    return 1;
}

/*-----------------------------------------------------------------------*/

static int do_stats_update(SysUserDataParams *params, const char *path)
{
    PRECOND(params != NULL, return 0);
    PRECOND(path != NULL, return 0);

    if (!stats) {
        DLOG("No statistics registered");
        return 0;
    }

    /* Don't bother updating the file if there are no changes to save. */
    int update_needed = stats_update_needed;

    /* Find and update each statistic given by the caller. */

    for (int i = 0; i < params->stat_values_len; i++) {
        int found = 0;
        for (int j = 0; j < num_stats; j++) {
            if (params->stat_values[i].id == stats[j].info.id) {
                found = 1;
                break;
            }
        }
        if (!found) {
            DLOG("Stat %u not found", params->stat_values[i].id);
            return 0;
        }
        if (isnan(params->stat_values[i].value)) {
            DLOG("Stat %u value is a NaN", params->stat_values[i].id);
            return 0;
        }
    }

    for (int i = 0; i < params->stat_values_len; i++) {
        int found = 0;
        int j;
        for (j = 0; j < num_stats; j++) {
            if (params->stat_values[i].id == stats[j].info.id) {
                found = 1;
                break;
            }
        }
        ASSERT(found);

        double real_value = 0;
        switch (stats[j].info.type) {
          case USERDATA_STAT_FLAG:
            real_value = (params->stat_values[i].value != 0) ? 1 : 0;
            if (stats[j].value && !real_value) {
                DLOG("Ignoring attempt to clear flag %u", stats[j].info.id);
                real_value = stats[j].value;
            }
            break;
          case USERDATA_STAT_UINT32:
          case USERDATA_STAT_UINT32_MAX:
            real_value = floor(params->stat_values[i].value);
            real_value = bound(real_value, 0, (double)0xFFFFFFFFU);
            break;
          case USERDATA_STAT_DOUBLE:
          case USERDATA_STAT_DOUBLE_MAX:
            real_value = params->stat_values[i].value;
            break;
        }
        if (stats[j].info.type == USERDATA_STAT_UINT32_MAX
         || stats[j].info.type == USERDATA_STAT_DOUBLE_MAX) {
            real_value = max(real_value, stats[j].value);
        }

        if (stats[j].value != real_value) {
            stats[j].value = real_value;
            update_needed = 1;
        }

    }  // for (int i = 0; i < params->stat_values_len; i++)

    /* If something changed, write out the updated data. */

    if (update_needed) {
        stats_update_needed = 1;
        const int result = save_stats_file(path);
        if (!result) {
            return 0;
        }
    }  // if (update_needed)

    stats_update_needed = 0;
    return 1;
}

/*-----------------------------------------------------------------------*/

static int do_stats_retrieve(SysUserDataParams *params, const char *path)
{
    PRECOND(params != NULL, return 0);
    PRECOND(path != NULL, return 0);

    if (!stats) {
        DLOG("No statistics registered");
        return 0;
    }

    /* Find and retrieve each requested statistic.  A value of zero is
     * returned for invalid IDs. */

    for (int i = 0; i < params->stat_buffer_len; i++) {
        int found = 0;
        for (int j = 0; j < num_stats; j++) {
            if (params->stat_buffer[i].id == stats[j].info.id) {
                params->stat_buffer[i].value = stats[j].value;
                found = 1;
                break;
            }
        }
        if (!found) {
            DLOG("Stat %u not found, setting default value 0",
                 params->stat_buffer[i].id);
            params->stat_buffer[i].value = 0;
        }
    }  // for (int i = 0; i < params->stat_buffer_len; i++)

    return 1;
}

/*-----------------------------------------------------------------------*/

static int do_stats_clear(SysUserDataParams *params, const char *path)
{
    PRECOND(params != NULL, return 0);
    PRECOND(path != NULL, return 0);

    if (!stats) {
        DLOG("No statistics registered");
        return 0;
    }

    if (unlink(path) != 0 && errno != ENOENT) {
        DLOG("unlink(%s) failed: %s", path, strerror(errno));
        return 0;
    }

    for (int i = 0; i < num_stats; i++) {
        stats[i].value = 0;
    }
    stats_update_needed = 0;

    return 1;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
